from typer import Option, Argument
import os
import subprocess
from pathlib import Path
from typing import List

from scriptz import cli, logger, Manager


@cli.command()
def run(paths: List[Path] = Argument(None), start: Path = Option(None)):
    """ Run scripts specified. """

    manager = Manager()
    paths = paths or [Path.cwd()]

    start = start and start.absolute()

    for script in manager.iterate_scripts(paths):
        if start and (script.path < start):
            logger.info(f"- Skipping: {script}")

        else:
            logger.info(f"+ Running: {script} ....................+")

            args = [script.handler.command] + script.handler.options
            args.append(script.path)

            env = manager.config.get_environ(script.path)

            process = subprocess.Popen(
                args,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                env=dict(os.environ, **env),
            )

            with process.stdout:
                log_subprocess_output(process.stdout)

            exit_code = process.wait()  # 0 means success
            logger.debug(f"Complete: Error Code = {exit_code}")


def log_subprocess_output(pipe):
    for line in iter(pipe.readline, b""):
        logger.info(line.decode().strip())



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Optional, Iterator

from . import Script

if TYPE_CHECKING:
    from . import Manager


class Folder:
    def __init__(self, manager: Manager, path: Path):
        self.manager = manager
        self.path = path

    def __str__(self):
        return self.name

    @property
    def relative_path(self):
        return self.path.relative_to(self.manager.root_dir)

    @property
    def name(self):
        return str(self.relative_path)

    @property
    def exists(self):
        return self.path.is_dir()

    @property
    def step_paths(self):
        return sorted(p for p in self.path.glob("[0-9]*_*"))

    @property
    def step_names(self):
        return [p.name for p in self.step_paths]

    @property
    def last_step_name(self) -> Optional[str]:
        last_step_name = self.step_names[-1:]
        if last_step_name:
            return last_step_name[0]

    @property
    def last_prefix(self) -> str:
        return (self.last_step_name or "_").split("_", maxsplit=1)[0]

    def get_next_prefix(self) -> str:
        next_prefix = int(self.last_prefix or -1) + 1
        return self.manager.config.zfill(next_prefix)

    def create_script(self, file_name: str, suffix: str):
        next_prefix = self.get_next_prefix()
        hold = self.path / f"{next_prefix}.hold"
        hold.touch(exist_ok=False)  # todo: real lock

        path = self.path / f"{next_prefix}_{file_name}{suffix}"
        self.manager.check(not path.exists(), f"Script {path} already exists.")
        path.touch(exist_ok=False)
        script = Script(manager=self.manager, path=path, folder=self)

        hold.unlink(missing_ok=False)  # todo: real lock

        return script

    def iterate_scripts(self) -> Iterator[Script]:
        for path in self.step_paths:
            if path.is_dir():
                folder = self.manager.get_folder(path=path)
                yield from folder.iterate_scripts()
            else:
                yield Script(manager=self.manager, path=path, folder=self)



from .environ import Environ
from .handler import Handler
from .script import Script
from .folder import Folder
from .config import Config
from .manager import Manager

__all__ = (
    "Config",
    "Environ",
    "Folder",
    "Handler",
    "Manager",
    "Script",
)



from configparser import ConfigParser
from pathlib import Path
from typing import Optional, Union

from . import Handler, Environ


class Config:
    def __init__(self, root_dir: Path):
        # get configuration values from scriptz.ini file
        self.ini_path = root_dir / "scriptz.ini"
        ini = self._read_ini_file(self.ini_path)
        cfg = "config"

        # maximum number of characters to keep when creating a file name
        self.file_length = max(ini.getint(cfg, "file_length", fallback=20), 6)

        # length of prefix used by str.zfill() to pad name with zeroes
        self.prefix_size = max(ini.getint(cfg, "prefix_size", fallback=2), 1)

        # handlers
        self.handlers = Handler.create_handlers(ini)
        self.file_types = list(h.display for h in self.handlers.values())
        self.default_type = self.file_types[0] if self.file_types else None
        self.file_types = list(dict.fromkeys(self.file_types))

        # secrets/environment variables
        self.env_path = root_dir / "secretz.ini"
        env_ini = self._read_ini_file(self.env_path, env=True)
        self.environs = Environ(env_ini=env_ini)

    def get_handler(self, key: Union[Path, str]) -> Optional[Handler]:
        if isinstance(key, Path):
            key = key.suffix
        return self.handlers.get(key)

    def get_environ(self, path: Union[Path, str]) -> dict:
        return self.environs.get(path=path)

    def zfill(self, number):
        if number == -1:
            return "9" * self.prefix_size
        return str(number).zfill(self.prefix_size)

    @property
    def is_valid(self):
        return self.handlers

    # private methods

    @classmethod
    def _read_ini_file(cls, ini_path: Path, env: bool = False) -> ConfigParser:
        ini = ConfigParser()
        if env:
            ini.optionxform = str

        if ini_path.is_file():
            ini.read_file(ini_path.open("r"))
        return ini



import logging
from rich.logging import RichHandler

logging.basicConfig(
    level="NOTSET",
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(rich_tracebacks=True)],
)
logger = logging.getLogger("rich")



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from . import Manager, Folder


class Script:
    def __init__(self, manager: Manager, path: Path, folder: Folder = None):
        self.manager = manager
        self.path = path
        self.folder = folder or self.manager.get_folder(path.parent)
        self.handler = manager.config.get_handler(path)

    def __str__(self):
        return f"{self.full_name} [{self.handler}]"

    @property
    def name(self):
        return self.path.name

    @property
    def full_name(self):
        return f"{self.folder.name}/{self.name}"



import os
import subprocess
import string
from pathlib import Path
from typer import Argument

from rich.prompt import Prompt

from scriptz import cli, Manager


@cli.command()
def new(folder_path: Path = Argument(None)):
    """ Create a new script. """
    manager = Manager()
    folder = manager.get_folder(folder_path)

    file_type = Prompt.ask(
        "File Type",
        choices=manager.config.file_types,
        default=manager.config.default_type,
    )
    handler = manager.config.get_handler(file_type)

    description = Prompt.ask("Description")

    file_name = make_file_name(description, manager.config.file_length)
    new_script = folder.create_script(file_name, handler.suffix)

    # initialize default content
    contents = [f"{handler.comment} {description}"]
    if handler.comment == "#":
        contents.insert(0, f"#!/usr/bin/env {handler.command}")
    new_script.path.write_text("\n".join(contents))

    editor = os.getenv("EDITOR", "vim")
    subprocess.call("%s %s" % (editor, new_script.path), shell=True)


def make_file_name(text: str, max_length: int = 20) -> str:
    max_length = max(max_length, 6)  # at least 6 characters
    text = text.lower()
    text = "".join(ch if ch in lower_alphanum else " " for ch in text)
    words = text.split()

    if len("_".join(words)) > max_length:
        filtered_words = list(word for word in words if word not in stop_words)
        if filtered_words:
            words = filtered_words

    trim = None
    text = "_".join(words[:trim])
    while len(text) > max_length:
        trim = (trim - 1) if trim is not None else -1
        text = "_".join(words[:trim])

    return text


lower_alphanum = set(string.ascii_lowercase + string.digits)

stop_words = {
    "a",
    "and",
    "as",
    "at",
    "but",
    "by",
    "for",
    "from",
    "i",
    "in",
    "it",
    "not",
    "of",
    "on",
    "or",
    "s",
    "t",
    "that",
    "the",
    "this",
    "to",
    "up",
    "with",
    "you",
}



from pathlib import Path

from scriptz import cli, Manager


@cli.command()
def init():
    """ Initialize current directory for scriptz. """
    manager = Manager()

    manager.check(
        not manager.ini_path.exists(),
        f"scriptz already initialized ({manager.ini_path.absolute()}).",
    )

    ini_content = (Path(__file__).parent.parent / "scriptz.ini").read_text()
    manager.ini_path.write_text(ini_content)

    manager.check(
        manager.ini_path.is_file(),
        error_msg=f"scriptz initialization ({manager.ini_path}) failed.",
        ok_msg=f"scriptz initialized ({manager.ini_path}) successfully.",
    )
